{# -*- engine: jinja -*- #}
{% for function_batch in all_functions | batch(10000) %}
<table class="listOfFunctions">
  <tr>
    <th>Function (File:Line)</th>
    <th>Call count</th>
    {% if not EXCLUDE_FUNCTION_COVERAGE %}
    <th>Line coverage</th>
    <th>Branch coverage</th>
    <th>Condition coverage</th>
    {% endif %}
    <th>Block coverage</th>
  </tr>
  {% for entry in function_batch %}
  <tr>
    {#- #}<td><a href="{%- if info.single_page %}#{{ entry['html_filename'] }}|
                       {%- else %}{{ entry['html_filename'] }}#
                       {%- endif %}l{{ entry['line'] }}">{{ entry["name"] }} ({{ entry["filename"] }}:{{ entry["line"] }})</a></td>
    {#- #}<td>
            {%- if entry["excluded"] %}excluded
            {%- else %}
            {%-  if entry["count"] == 0 %}not called
            {%-  else %}called {{ entry["count"] }} time{% if entry["count"] > 1 %}s{% endif %}
            {%-  endif -%}
            {%- endif -%}
    {#- #}</td>
    {%- if not EXCLUDE_FUNCTION_COVERAGE %}
    {#- #}<td>{{ entry["line_coverage"] }}%</td>
    {#- #}<td>{{ entry["branch_coverage"] }}%</td>
    {#- #}<td>{{ entry["condition_coverage"] }}%</td>
    {%- endif %}
    {#- #}<td>{{ entry["blocks"] }}%</td>
    {#- -#}
  </tr>
  {% endfor %}
</table>
{% endfor %}
