{# -*- engine: jinja -*- #}
<div class="filelist Box m-3">
    <div class="Box-header d-flex flex-space-between">
        <div role="gridcell" class="mr-3 flex-shrink-0" style="width: 16px;">
        </div>
        <div role="rowheader" class="flex-auto min-width-0 col-md-2 mr-3">
        </div>
        <div role="gridcell" class="flex-auto min-width-0 col-5 mr-3">
        </div>
        {% if not EXCLUDE_CONDITIONS %}
        <div role="gridcell" class="color-fg-muted d-flex flex-justify-end flex-wrap col-2" >
        </div>
        {% endif %}
        {% if SHOW_DECISION %}
        <div role="gridcell" class="color-fg-muted d-flex flex-justify-end flex-wrap col-2">
        </div>
        {% endif %}
        {% if not EXCLUDE_CALLS %}
        <div role="gridcell" class="color-fg-muted d-flex flex-justify-end flex-wrap col-2" >
        </div>
        {% endif %}
    </div>
    {% for row in entries %}
    <div class="Box-row Box-row--focus-gray py-2 d-flex position-relative">
        <div role="rowheader" class="flex-auto min-width-0 mr-3 col-md-2">
            <div class="d-flex">
            <span class="mr-2">
            <svg aria-hidden="true" focusable="false" role="img" class="color-fg-muted" viewBox="0 0 16 16" width="16" height="16" fill="currentColor" style="display: inline-block; user-select: none; vertical-align: text-bottom; overflow: visible;"><path d="M2 1.75C2 .784 2.784 0 3.75 0h6.586c.464 0 .909.184 1.237.513l2.914 2.914c.329.328.513.773.513 1.237v9.586A1.75 1.75 0 0 1 13.25 16h-9.5A1.75 1.75 0 0 1 2 14.25Zm1.75-.25a.25.25 0 0 0-.25.25v12.5c0 .138.112.25.25.25h9.5a.25.25 0 0 0 .25-.25V6h-2.75A1.75 1.75 0 0 1 9 4.25V1.5Zm6.75.062V4.25c0 .138.112.25.25.25h2.688l-.011-.013-2.914-2.914-.013-.011Z"></path></svg>
            </span>
            <span class="css-truncate css-truncate-target d-block width-fit">
            {% if row.link is not none %}
            <a class="" title="{{row.filename}}" href="{% if info.single_page %}#{% endif %}{{row.link}}">{{row.filename}}</a>
            {% else %}
            {{row.filename}}
            {% endif %}
            </span>
            </div>
        </div>
        <div role="gridcell" class="flex-auto min-width-0 r-3 d-flex flex-justify-end flex-wrap flex-items-baseline">
            <span class="text-small color-fg-muted mr-2">{{row.lines.exec}} of {{row.lines.total}} lines <span class="line-percentage-text {% if SHOW_DECISION and not EXCLUDE_CALLS %}d-none{% endif %}">({{row.lines.coverage}}%)</span></span>
            <div class="Label--minw meter-label {% if SHOW_DECISION and not EXCLUDE_CALLS %}d-block{% else %}d-none{% endif %}">
                <span class="Label mr-1 Label--{{row.lines.class}}">{{row.lines.coverage}}%</span>
            </div>
            {% if row.lines.coverage == "-" %}
            <span style="width: 200px;"></span>
            {% else %}
            <span class="meter Progress Progress--large d-inline-flex {% if SHOW_DECISION and not EXCLUDE_CALLS %}d-none{% endif %}" style="width: 200px;">
                <span class="Progress-item {{row.lines.class}}" style="width: {{row.lines.coverage}}%"></span>
            </span>
            {% endif %}
        </div>
        <div role="gridcell" class="color-fg-muted d-flex flex-justify-end flex-wrap col-2">
            <span class="text-small color-fg-muted mr-2">{{row.functions.exec}} of {{row.functions.total}} functions</span>
            <div class="Label--minw">
                <span class="Label mr-1 Label--{{row.functions.class}}">{{row.functions.coverage}}%</span>
            </div >
        </div>
        <div role="gridcell" class="color-fg-muted d-flex flex-justify-end flex-wrap col-2">
            <span class="text-small color-fg-muted mr-2">{{row.branches.exec}} of {{row.branches.total}} branches</span>
            <div class="Label--minw">
                <span class="Label mr-1 Label--{{row.branches.class}}">{{row.branches.coverage}}%</span>
            </div>
        </div>
        {% if not EXCLUDE_CONDITIONS %}
        <div role="gridcell" class="color-fg-muted d-flex flex-justify-end flex-wrap col-2">
            <span class="text-small color-fg-muted mr-2">{{row.conditions.exec}} of {{row.conditions.total}} conditions</span>
            <div class="Label--minw">
                <span class="Label mr-1 Label--{{row.conditions.class}}">{{row.conditions.coverage}}%</span>
            </div>
        </div>
        {% endif %}
        {% if SHOW_DECISION %}
        <div role="gridcell" class="color-fg-muted d-flex flex-justify-end flex-wrap col-2">
            <span class="text-small color-fg-muted mr-2">{{row.decisions.exec}} of {{row.decisions.total}} decisions</span>
            <div class="Label--minw">
                <span class="Label mr-1 Label--{{row.decisions.class}}">{{row.decisions.coverage}}%</span>
            </div>
        </div>
        {% endif %}
        {% if not EXCLUDE_CALLS %}
        <div role="gridcell" class="color-fg-muted d-flex flex-justify-end flex-wrap col-2">
            <span class="text-small color-fg-muted mr-2">{{row.calls.exec}} of {{row.calls.total}} calls</span>
            <div class="Label--minw">
                <span class="Label mr-1 Label--{{row.calls.class}}">{{row.calls.coverage}}%</span>
            </div>
        </div>
        {% endif %}
    </div>
  {% endfor %}
</div>

