/*** Summary formats ***/

.summary
{
    display: flex;
    flex-flow: row wrap;
    max-width: 100%;
    justify-content: flex-start;
}

.summary > table
{
    flex: 1 0 7em;
    border: 0;
}

.summary > :last-child {
    margin-left: auto;
}

table.legend
{
    color: black;
    display: flex;
    flex-flow: row wrap;
    justify-content: flex-start;
}

table.legend th[scope=row]
{
    font-weight: normal;
    text-align: right;
    white-space: nowrap;
}

table.legend td
{
    color: blue;
    text-align: left;
    white-space: nowrap;
    padding-left: 5px;
}

table.legend td.legend
{
    color: black;
    font-size: 80%;
}

table.legend td.warning_text
{
    color: var(--warning_color);
}

table.coverage td,
table.coverage th
{
    text-align: right;
    color: black;
    font-weight: normal;
    white-space: nowrap;
    padding-left: 5px;
    padding-right: 4px;
}

table.coverage td
{
    background-color: LightSteelBlue;
}

table.coverage th[scope=row]
{
    color: black;
    font-weight: normal;
    white-space: nowrap;
}

table.coverage th[scope=col]
{
    color: blue;
    font-weight: normal;
    white-space: nowrap;
}

table.legend span
{
    margin-right: 4px;
    padding: 2px;
}

table.legend span.coverage-unknown,
table.legend span.coverage-none,
table.legend span.coverage-low,
table.legend span.coverage-medium,
table.legend span.coverage-high
{
    padding-left: 3px;
    padding-right: 3px;
}

table.legend span.coverage-unknown,
table.coverage td.coverage-unknown,
table.file-list td.coverage-unknown
{
    background-color: var(--unknown_color) !important;
}

table.legend span.coverage-none,
table.legend span.coverage-low,
table.coverage td.coverage-none,
table.coverage td.coverage-low,
table.file-list td.coverage-none,
table.file-list td.coverage-low
{
    background-color: var(--low_color) !important;
}

table.legend span.coverage-medium,
table.coverage td.coverage-medium,
table.file-list td.coverage-medium
{
    background-color: var(--medium_color) !important;
}

table.legend span.coverage-high,
table.coverage td.coverage-high,
table.file-list td.coverage-high
{
    background-color: var(--high_color) !important;
}
/*** End of Summary formats ***/
